/******************************************************************************
                                                    
 amalgamator.cpp


        This program, the "amalgamator," reads input from UNIX sockets, which
          it expects to consist of video frames, and then processes that data
          for display in a single buffer. The input images are arranged in a
          user-specified fashion, and composited onto a single large buffer.
          This format is due to the program's intended use with SAGE, which 
          requires, for simplicity, a raw frame buffer.

        Currently, the amalgamator is being modified to read input from network
          sockets (the major change from v3.0 to v4.0). It will be integrated
          with code that reads and decodes XVid data, so that VLC will not be
          needed.

                              
******************************************************************************/


//#define TIMING




#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <vector>
#include <errno.h>

#ifdef TIMING
#include <sys/time.h>
#endif


#include <pthread.h>


#include "Vector.h"

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>


//network headers
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>


//xvid headers
#include "vvxvid.h"

//tile rendering library header
#include "tr.h"


#ifdef SAGE_APP
// SAGE stuff
#include "sail.h"
#include "misc.h"
#endif

#define X_SLIDE_STEP 5
#define Y_SLIDE_STEP 3

#define BYTES_PER_PIXEL 3

#define PRINTF(x) NULL//x

#define INDIVIDUAL_PIXEL_AT(img, x,y) (Vector4(img[((y*individual_width)+x)*4],img[((y*individual_width)+x)*4+1], img[((y*individual_width)+x)*4+2], img[((y*individual_width)+x)*4+3]))  


using namespace std;


class point { 
public: 
    point() {x=0;y=0;}
    point(int a, int b) { x=a; y=b; }

    int x; 
    int y; 
};


// ------------GLOBALS--------------- //

#ifdef SAGE_APP
sail sageInf;   // SAIL object
#endif


#ifdef USE_SHADER_PROGRAM
#include <dlfcn.h>
#include <Cg/cg.h>
#include <Cg/cgGL.h>
CGcontext   cgContext;
CGprogram   cgProgram;
CGprofile   cgFragmentProfile;
CGparameter cg_imwidth, cg_imheight, cg_focal_length;
#endif

#define TILE_WIDTH 1920
#define TILE_HEIGHT 1080
//#define TILE_WIDTH 860
//#define TILE_HEIGHT 540
//#define TILE_WIDTH 512
//#define TILE_HEIGHT 512
#define BORDER 0

TRcontext *tr;

//there is one mutex per individual framebuffer. In order to read from or write to that buffer,
//the operator must hold the corresponding lock!!
pthread_mutex_t **buffer_mutexes = 0;

// global variables for image sizes with arbitrary initial values;
// these values should be overwritten with the correct input that
// is parsed from the command line arguments to the program
int total_width = 1920;
int total_height = 1080;
int individual_width = 1920;
int individual_height = 1080;
int number_of_streams = 1;
int base_port_number = -1;         //this will contain the port number on which the amalgamator can expect to find
                                //the first stream. Subsequent frames (number_of_streams, to be exact) will be
                                //received on subsequent port numbers. 
                                //(e.g. base = 1234, 4 streams =>
                                //  stream 0 is expected on port 1234
                                //  strean 1 is expected on port 1235
                                //  stream 2 is expected on port 1236
                                //  stream 3 is expected on port 1237)
GLuint *texture = 0;

// this will hold the position of each frame in the final buffer.
// its size will be 2*number_of_streams. the even entries will contain
// the starting x position for that stream, and the odds will contain the y.
int *positions = 0;

int total_individual_bytes = individual_width*individual_height*BYTES_PER_PIXEL;

#ifdef TIMING
struct timeval tv;
#endif


// global variables for the image buffers themselves
unsigned char *total_buffer = 0;
unsigned char **individual_buffers = 0;   // eventually, this should not exist, and
                                // 'individual_buffers' should map directly
                                  // into total_buffer.
//Used by socket threads to flag when it gets a bitrate from the sender
int * brflag = 0;
//Used by socket threads to store bitrate from the sender
unsigned int * streambr = 0;
//Array of socket ids for thread communication on bandwidth change
int * socketfd = 0;
//Global target bitrate for all sockets
unsigned int globalbr = 0;
  

unsigned char *tempReadBuffer = 0;

int largest_fd = 0;             //I think we need this to be able to select() from all of our open
                                //file descriptors (in our case, these will all be local unix sockets)
int *read_sockets = 0;
int *write_sockets = 0;
float focal_length = 3180;


bool stitching_snapshot = 0;                //will be set to 1 when a snapshot is requested for stitching purposes


int cur_img_num = 0;


FILE *stdin_ptr = 0;

socklen_t socklen = sizeof( struct sockaddr_in );


// ------------ END GLOBALS --------------- //

// ------------ PROTOTYPES --------------- //


void error(char *);
int* init(int argc, char *argv[], int *sockfd);
void parse_arguments(int argc, char **argv);
void uninit();

void drawPlane(const Matrix4x4 &mTransform);
void Draw(void);
void socketInit();

void StitchImages();
void *stream_thread_main(void *args);
bool Initialize (int argc, char **argv);

// ------------ END PROTOTYPES ------------//


bool Initialize (int argc, char **argv)
{
    glClearColor (0.0f, 0.0f, 0.0f, 0.0f);
    
    //begin parsing inputs
    if(argc < 5)
    {
        fprintf(stderr, "Too few arguments.\n");
        fprintf(stderr, "Expected usage:\n    ./amalgamator N TW TH IW IH x y x y ...\n\
         Where N is the number of input streams, TW and TH are the width\n\
         and height of the final composite image, IW and IH are the width\n\
         and height of the individual images, and each pair of 'x' and 'y'\n\
         correspond to the starting coordinates of each stream, in order.\n");
        exit(0);
    }

    number_of_streams = atoi(argv[1]);

    if( argc < 5 + number_of_streams*2 )
    {
        fprintf(stderr, "Too few arguments.\n");
        fprintf(stderr, "You need a pair of starting coordinates for each stream.\n");
        fprintf(stderr, "Example: ./amalgamator 2 640 960 640 480 0 0 0 480 will tile\n");
        fprintf(stderr, "640x480 screens vertically.\n");
        exit(0);
    }

    individual_width = atoi(argv[4]);
    individual_height = atoi(argv[5]);

    total_individual_bytes = individual_width*individual_height*BYTES_PER_PIXEL;

    //parse initial positions for each image
    positions = new int[number_of_streams*2];

    for(int i = 0; i < number_of_streams*2; i++)
    {
        positions[i] = atoi( argv[6+i] );
    }
  
//1234
    brflag = new int[1];
    brflag[0] = 0;

    socketfd = new int[number_of_streams];
    for(int i = 0; i < number_of_streams; i++)
    {
        socketfd[i] = 0;
    }

    streambr = new unsigned int[number_of_streams];
    for(int i = 0; i < number_of_streams; i++)
    {
        streambr[i] = 0;
    }
  
    //initialize the individual stream buffers
    individual_buffers = new unsigned char*[number_of_streams];
    for(int i = 0; i < number_of_streams; i++)
    {
        individual_buffers[i] = new unsigned char[individual_width*individual_height*BYTES_PER_PIXEL];
    }


    //allocate space for the total buffer
    total_buffer = new unsigned char[total_width * total_height * BYTES_PER_PIXEL];

    

    for(int i=0;i<individual_width*individual_height*BYTES_PER_PIXEL;i++)
        individual_buffers[0][i] = (i%3 == 0 ? 64 : 0);
    
    if(number_of_streams > 1)
        for(int i=0;i<individual_width*individual_height*BYTES_PER_PIXEL;i++)
            individual_buffers[1][i] = (i%3 == 2 ? 64 : 0);
        
    texture = new GLuint[number_of_streams];
    //initialize OpenGL textures for the streams
    for(int i = 0; i < number_of_streams; i++)
    {
        glGenTextures(1, &texture[i]);
        glBindTexture(GL_TEXTURE_2D, texture[i]);
        if(BYTES_PER_PIXEL == 3)
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, individual_width, individual_height, 0, GL_RGB, GL_UNSIGNED_BYTE, individual_buffers[i]); 
        else
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, individual_width, individual_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, individual_buffers[i]); 
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); 
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
       
        glEnable(GL_TEXTURE_2D);
    }


#if 0

    //-------------------------------------
    //------INITIALIZE SOCKETS-------------
    //-------------------------------------

    int servlen, clilen;
    struct sockaddr_un cli_addr, serv_addr;

    read_sockets = new int[number_of_streams];
    write_sockets = new int[number_of_streams];

    // open the sockets.....

    for(int i = 0; i < number_of_streams; i++)
    {
        char *tempName = new char[25];      // will hold: "vlc_stream_socket_xxxxxx"
        char *tempNumber = new char[10];    //'10' is just a number large enough for the digits 
        sprintf(tempNumber, "%d", i);
        strncpy(tempName, "vlc_stream_socket_", 18);
        strcpy(tempName+18, tempNumber);

        if ((read_sockets[i] = socket(AF_UNIX,SOCK_STREAM,0)) < 0)
        {
            fprintf(stderr, "Error creating socket number %d.\n", i);
            exit(1);
        }
        bzero((char *) &serv_addr, sizeof(serv_addr));
        serv_addr.sun_family = AF_UNIX;
        //by this point, tempName should be an array of characters holding
        //a string of the form "vlc_stream_socket_xxxxxx," where xxxxxx is
        //replaced by a number for each socket.
        strcpy(serv_addr.sun_path, tempName);
        servlen=strlen(serv_addr.sun_path) + 
                            sizeof(serv_addr.sun_family);
        if(bind(read_sockets[i],(struct sockaddr *)&serv_addr,servlen)<0)
        {
            //the socket might already exist, so try removing it and binding again...
            remove(tempName);
            if(bind(read_sockets[i] , (struct sockaddr *)&serv_addr , servlen ) < 0 )
            {
                fprintf(stderr, "Error binding socket number %d.\n", i);
                exit(1);
            }
        }
        printf("Opened socket %s.\n", tempName); 
        delete(tempName);    
        delete(tempNumber);
    }

    //Initialize the connections and get a writable socket.
    for(int i = 0; i < number_of_streams; i++)
    {
        if(listen(read_sockets[i],5))
        {
            fprintf(stderr, "Error listening on read_sockets[%d].\n", i);
            exit(1);
        }

        clilen = sizeof(cli_addr);
        write_sockets[i] = accept(read_sockets[i],(struct sockaddr *)&cli_addr, (socklen_t*)&clilen);
        if (write_sockets[i] < 0) 
        {
            fprintf(stderr, "Error accepting on write_sockets[%d].\n", i);
            exit(1);
        }
        
        printf("read_sockets[%d]: %d. write_sockets[%d]: %d.\n", i, read_sockets[i], i, write_sockets[i]);
    }
   
    //I'm pretty sure this code/comment section existed because I was unclear as to the usage of
    //the select function. Kept for...posterity? Maybe?
    
    // in case file descriptors aren't allocated in order
    // (...but is that even possible?) 
    // I believe we need to keep track of the largest socket to 'select' from all of ours...
    // or something like that?
    //for(int i = 0; i < number_of_streams; i++)
    //    if(write_sockets[i] > largest_fd)
    //        largest_fd = write_sockets[i];

    printf("Sockets initialized.\n");

#endif

    //-------------------------------------
    //------INITIALIZE SAGE----------------
    //-------------------------------------
    

    #ifdef SAGE_APP
    
    printf("Initializing SAGE...");

    
    // initialize SAGE...
    sageRect imageMap;
    imageMap.left = 0.0;
    imageMap.right = 1.0;
    imageMap.bottom = 0.0;
    imageMap.top = 1.0;

    sailConfig scfg; 
    scfg.cfgFile = strdup("sage.conf");
    scfg.appName = strdup("amalgamator");
    scfg.execConfig = strdup("two_horiz_HD");
    scfg.rank = 0;
    scfg.ip = NULL;    
    scfg.resX = total_width;
    scfg.resY = total_height;
    scfg.imageMap = imageMap;

    //TODO: CHECK THIS!!! OpenGL can take RGBA textures and just render an RGB texture... is this what we want? This 
    //would decrease the bandwidth sent to SAGE.... which would certainly be a good thing.


    if(BYTES_PER_PIXEL == 3)
    {
        scfg.colorDepth = 24;
        scfg.pixFmt = TVPIXFMT_888;
    } else {
        scfg.colorDepth = 32;
        scfg.pixFmt = TVPIXFMT_8888;
    }
    scfg.rowOrd = BOTTOM_TO_TOP;
    scfg.nwID = 1;

    sageInf.init(scfg);

    printf("done.\n");    
    
    #endif
    
    //-------------------------------------
    //------INITIALIZE SHADER--------------
    //-------------------------------------
    
    #ifdef USE_SHADER_PROGRAM

    //Get the function pointers from libGL.so. I have no idea why this won't work automatically...
    void *dynLib = dlopen("libGL.so", RTLD_NOW);
    void (*glGenProgramsARB)(GLsizei, GLuint*) = (void (*)(GLsizei,GLuint*)) dlsym(dynLib, "glGenProgramsARB");
    void (*glBindProgramARB)(GLenum, GLuint) = (void (*)(GLenum, GLuint)) dlsym(dynLib, "glBindProgramARB");
    //void (*glProgramStringARB)(GLenum, GLenum, GLsizei, const GLvoid*) = (void (*)(GLenum, GLenum, GLsizei, const GLvoid*)) dlsym(dynLib, "glProgramStringARB");

    glEnable(GL_FRAGMENT_PROGRAM_ARB);

    if(glGenProgramsARB == NULL)
        printf("glGenProgramsARB is null!\n");
    if(glBindProgramARB == NULL)
        printf("glBindProgramARB is null!\n");

    unsigned int shader_num = 0;
    glGenProgramsARB(1, &shader_num);
    glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, shader_num);

    //Cg initialization code
    cgContext = cgCreateContext();
    cgProgram = cgCreateProgramFromFile(cgContext, CG_SOURCE, "spherical_warp.cg", CG_PROFILE_ARBFP1, "main", (const char **)argv);
    fprintf(stderr, "cgProgram: %d. isProgram? %s.\n", cgProgram, cgIsProgram(cgProgram) ? "yes" : "no");
    cgGLLoadProgram(cgProgram); 
    cgGLEnableProfile(CG_PROFILE_ARBFP1);
    cgGLBindProgram(cgProgram);
    
    //disable the shader program by default just to be safe;
    //it'll be reenabled when we draw something.
    glDisable(GL_FRAGMENT_PROGRAM_ARB);
   
    #endif


    //--------------------------------------
    //-----INITIALIZE TILE RENDERER---------
    //--------------------------------------


    #ifdef TILE_RENDER

    /* Setup tiled rendering */
    tr = trNew();
    trTileSize(tr, TILE_WIDTH, TILE_HEIGHT, BORDER);
    /* We don't call trTileBuffer() since we're not interested in getting
    * each individual tile's data.
    */
    trImageSize(tr, total_width, total_height);
    if(BYTES_PER_PIXEL == 3)
        trImageBuffer(tr, GL_RGB, GL_UNSIGNED_BYTE, total_buffer);
    else
        trImageBuffer(tr, GL_RGBA, GL_UNSIGNED_BYTE, total_buffer);
                                                                             
    #endif

    //--------------------------------------
    //-----INITIALIZE MUTEX VARIABLES-------
    //--------------------------------------
    
    buffer_mutexes = new pthread_mutex_t*[number_of_streams];
    for(int i = 0; i < number_of_streams; i++)
    {
        buffer_mutexes[i] = new pthread_mutex_t();
        if(pthread_mutex_init(buffer_mutexes[i], NULL))
        {
            fprintf(stderr, "Error! Unable to initialize buffer mutex %d.\n", i);
            exit(1);
        }
    }
        

    return true;														// Return true (Initialization Successful)
}

void Deinitialize (void)    											// Any User DeInitialization Goes Here
{
    // Destroy Our Cg Context And All Programs Contained Within It
    delete [] total_buffer;
    //trDelete(tr);
#ifdef TILE_RENDER    
    trDelete(tr);
#endif

}


int brindex = 0;
unsigned char bwnum[30];
void Update (int key,int x,int y)    									// Perform Motion Updates Here
{
  
    // A 'b' keystroke by the user gives the prompt to enter a new target bitrate
    if(key == 'b' || key == 'B')
    {
       fprintf(stderr, "Enter Bit Rate: ");
       brindex = 0;
       return;
    }

    // A numeric keystroke reads into the bandwidth number buffer
    if(key >= 48 && key <= 57)
    {
       fprintf(stderr, "%c", (char)key);
       bwnum[brindex] = (char)key;
       if(brindex < 27)
          brindex++;
       return;
    }

    // An 'r' keystroke locks in the new target bandwidth using the bwnum buffer
    if(key == 'r' || key == 'R')
    {
       int streamrate;
       if(brindex == 0)
          return;
       bwnum[brindex] = '\0';
       globalbr = atoi((const char *)bwnum);
       // calculate target bitrate per stream
       streamrate = globalbr / number_of_streams;
       fprintf(stderr, "\n");
       // send target bitrate to senders using array of socket ids
       for(int i = 0; i < number_of_streams; i++)
       {
          send(socketfd[i], &streamrate, 4, 0);	    
       }
       fprintf(stderr, "\n Global Bit Rate changed to %d, %d for each stream.\n", globalbr * 1000, streamrate * 1000);
       brindex = 0;
       return;
    }

    // Used internally by the threads to the total bitrate from the senders
    if(key == 'p' || key == 'P')
    {
       if(brflag[0] == 0)
          return;
       int tempbr = 0;
       //total bitrates from all streams
       for(int i = 0; i < number_of_streams; i++)
       {
          tempbr += streambr[i];
          streambr[i] = 0;
       }
       fprintf(stderr, "Bit Rate: %d\n", tempbr);
       brflag[0] = 0;
       return;
    }
   

    if ( key==27 )									// Is ESC Being Pressed?
    {
    	Deinitialize();
    	exit(0);							// Terminate The Program
    }
    
    if(key == 'w' || key == 'W')
        positions[2*cur_img_num+1] += (key == 'w' ? 1 : 3);
    if(key == 's' || key == 'S')
        positions[2*cur_img_num+1] -= (key == 's' ? 1 : 3);
    if(key == 'a' || key == 'A')
        positions[2*cur_img_num] -= (key == 'a' ? 1 : 3);
    if(key == 'd' || key == 'D')
        positions[2*cur_img_num] += (key == 'd' ? 1 : 3);

    if(key == 'n' || key == 'N')
        cur_img_num = (cur_img_num + 1) % number_of_streams;

    if(key == 'e' || key == 'E')
        focal_length += 20;
    if(key == 'c' || key == 'C')
        focal_length -= 20;

    if(key == 'q' || key == 'Q')
        StitchImages();

    fprintf(stderr, "focal length: %f.\n", focal_length);

    glutPostRedisplay();
}


//Stitches all of the images in the global variable ppms[].
//Makes assumptions that the sizes of the images are all the same.
//See the README for more information.
void StitchImages()
{
        
    //the step is the interval at which the images are 'slid' across each other.
    int x_step = 5;
    int y_step = 3;

    //ppms[0]->x_disp = ppms[0]->y_disp = 0;
    positions[0] = positions[1] = 0;

    printf("Stitching......\n");

    //the algorithm compares images piecewise, in pairs.
    //the image stitching algorithm generates a random number of points in a region
    //of suspected overlap, and for each possible transformation of the second image
    //determines the error between the corresponding points in the first image with 
    //a simple RGB error test.
    int num_of_random_points = 2000;

    for(int current_fixed_image = 0 ; current_fixed_image < number_of_streams- 1 ; current_fixed_image++)
    {
        
        printf("-----Stitching images %d and %d.\n", current_fixed_image, current_fixed_image+1);
    
        printf("-----Generating %d random points...", num_of_random_points);
        
        vector<point> random_points;
        for(int i = 0; i < num_of_random_points; i++)
        {
            //generate random points in the window:
            // x from width/2.0 to width
            // y from width/4.0 to 3.0/4.0 * height

            random_points.push_back( point((int)((rand() / (float)RAND_MAX) * (individual_width/2.0) + individual_width/2.0) , (int)(((rand() / (float)RAND_MAX) * (individual_height/2.0) + individual_height/4.0))) );
           
    
        }
        printf("done.\n-----Determining optimal shift...");    

    

        int optimal_x_shift = 0, optimal_y_shift = 0;
        double lowest_error = -1;
        
        int min_x_shift =  (int)( individual_width/2.0);
        int max_x_shift =  (int)( individual_width);
        int min_y_shift =  (int)(-individual_height/20.0);
        int max_y_shift =  (int)( individual_height/20.0);
                
    
        for(int current_x_shift = min_x_shift; current_x_shift < max_x_shift; current_x_shift+=x_step)
        {
            for(int current_y_shift = min_y_shift; current_y_shift < max_y_shift; current_y_shift+=y_step)
            {
                //int current_y_shift = 0;
                //iterate over the valid pixels of the overlapping regions of the pictures
                //and calculate the error for that particular transformation, keeping track of
                //the 'x' and 'y' shift for the transformation with a minimal error
                float current_error = 0;
                int num_of_pixels_used = 0;
                Vector4 temp1, temp2;



                //TODO!!! Check this code for alpha compatibility!!


                
                for(int i = 0 ; i < random_points.size() ; i++)
                {
                    //check whether, using the current transformation, the 
                    // given point is both in the input image and the output
                    // image, and that the point has full alpha in both.
                    if(random_points.at(i).x-current_x_shift >= 0 && random_points.at(i).x-current_x_shift < individual_width && random_points.at(i).y-current_y_shift >= 0 && random_points.at(i).y-current_y_shift < individual_height)// && (individual_buffers[current_fixed_image][(random_points.at(y)*individual_width+random_points.at(x))*BYTES_PER_PIXEL+3] > 0.001) && individual_buffers[current_fixed_image][((random_points.at(y)-current_y_shift)*individual_width+random_points.at(x)-current_x_shift)*BYTES_PER_PIXEL+3] > 0.001)
                    {
                        //if so, use the pixel, and increment the error.
                        num_of_pixels_used++;
                        temp1 = Vector4(INDIVIDUAL_PIXEL_AT(individual_buffers[current_fixed_image], random_points.at(i).x, random_points.at(i).y));
                        
                        //individual_buffers[current_fixed_image][IMG_AT(random_point.at(i).x, random_point.at(i).y)]);//modifiedppms[current_fixed_image]->GetPixel( random_points.at(i).x , random_points.at(i).y );
                        temp2 = Vector4(INDIVIDUAL_PIXEL_AT(individual_buffers[current_fixed_image], random_points.at(i).x-current_x_shift, random_points.at(i).y-current_y_shift)); //Vector4(individual_buffers[current_fixed_image][IMG_AT(random_points.at(i).x - current_x_shift , random_points.at(i).y - current_y_shift );
                        current_error += (fabs(temp1.x - temp2.x) + fabs(temp1.y - temp2.y) + fabs(temp1.z - temp2.z)) * (fabs(temp1.x - temp2.x) + fabs(temp1.y - temp2.y) + fabs(temp1.z - temp2.z));

                    }

                } 

                //after you have the total error for this transformation, check to see it it's
                 //a new low. If so, keep track of it; if not, move on.
                if(num_of_pixels_used == 0 && lowest_error < 0)
                    printf("num used == 0, error < 0!\n"); 
                if( (current_error / num_of_pixels_used <= lowest_error || lowest_error < 0) && num_of_pixels_used != 0)
                {
                    lowest_error = current_error / num_of_pixels_used;
                    optimal_x_shift = current_x_shift;
                    optimal_y_shift = current_y_shift;
                }
            }
        }

    
        printf("Match found!\n");
        printf("-----Optimal x-shift: %d. Optimal y-shift: %d. Error:%f.\n", optimal_x_shift, optimal_y_shift, lowest_error);
        //modifiedppms[current_fixed_image+1]->x_disp = optimal_x_shift; modifiedppms[current_fixed_image+1]->y_disp = optimal_y_shift;
        positions[(current_fixed_image+1)*2] = optimal_x_shift; positions[(current_fixed_image+1)*2+1] = optimal_y_shift;


        printf("-----Finished stitching images %d and %d.\n", current_fixed_image, current_fixed_image+1);

    }

    //actually update the displacements
    for(int i = 0 ; i < number_of_streams ; i++)
    { 
        if(i)
        {
            //ppms[i]->x_disp = modifiedppms[i]->x_disp + ppms[i-1]->x_disp;  
            positions[i*2] += positions[(i-1)*2];
            //ppms[i]->y_disp = modifiedppms[i]->y_disp + ppms[i-1]->y_disp; 
            positions[i*2+1] += positions[(i-1)*2+1];
        } else {
            //ppms[i]->x_disp = modifiedppms[i]->x_disp;
            //ppms[i]->y_disp = modifiedppms[i]->y_disp;
        }
    }

    //printf("-----Blending images...");
//    if(blending)
  //      BlendImages();
    //printf("done.\n");

    printf("Finished stitching all images.\n\n\n");
}












void Draw (void)
{


    PRINTF(printf("Beginning of the draw function.\n"));


    #ifdef SAGE_APP
    #ifndef TILE_RENDER
        glDrawBuffer(GL_AUX0);
    #endif
    #endif

    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity ();
    glMatrixMode( GL_MODELVIEW);
    glLoadIdentity();

    memset(total_buffer, 0, total_width*total_height*BYTES_PER_PIXEL);


#ifdef TILE_RENDER
    trOrtho(tr, 0, total_width, 0, total_height, 0.001, FLT_MAX);
#else 
    gluOrtho2D(0,total_width,0,total_height);
#endif

#ifdef TIMING
    gettimeofday(&tv, NULL);
#endif

#ifdef TILE_RENDER    
    do {
        trBeginTile(tr);
#endif


        //for(int i = 0; i < number_of_streams; i++)
        for(int i = number_of_streams-1; i >= 0; i--)
        {
            #ifdef USE_SHADER_PROGRAM
    
            cg_imwidth = cgGetNamedParameter(cgProgram, "imwidth");
            cg_imheight = cgGetNamedParameter(cgProgram, "imheight");
            cg_focal_length = cgGetNamedParameter(cgProgram, "focal_length");
            cgGLSetParameter1f(cg_imwidth, individual_width);
            cgGLSetParameter1f(cg_imheight, individual_height);
            cgGLSetParameter1f(cg_focal_length, focal_length);

            glEnable(GL_FRAGMENT_PROGRAM_ARB);

            #endif
        
            glBindTexture(GL_TEXTURE_2D, texture[i]);


        while(pthread_mutex_lock(buffer_mutexes[i]))
        {
            usleep(100);
            //fprintf(stderr, "Error! Unable to lock buffer mutex %d.\n", i);
            //exit(1);
        }

            //for(int j = 0; j < individual_width*individual_height*BYTES_PER_PIXEL; j++)
            //    individual_buffers[i][j] = 1;

            //printf("Some VALUEZZZ: (%d, %d, %d, %d)\n", individual_buffers[i][800], individual_buffers[i][801], individual_buffers[i][802], individual_buffers[i][803]);
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

        //glBlendFunc(GL_DST_ALPHA, GL_ONE_MINUS_DST_ALPHA);

            if(BYTES_PER_PIXEL == 3)
                glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, individual_width, individual_height, 0, GL_RGB, GL_UNSIGNED_BYTE, individual_buffers[i]);   
            else
                glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, individual_width, individual_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, individual_buffers[i]); 


            
            Matrix4x4 ex = Matrix4x4();
            ex.MakeTranslate(positions[2*i],positions[2*i+1],0);
            //fprintf(stderr, "positions: (%d, %d), for img: %d.\n", positions[2*i], positions[2*i+1], i);
            drawPlane(ex);
  //          glFinish();
  //          glutPostRedisplay();

        while(pthread_mutex_unlock(buffer_mutexes[i]))
        {
            usleep(100);
            //fprintf(stderr, "Error! Unable to unlock buffer mutex %d.\n", i);
            //exit(1);
        }


            #ifdef USE_SHADER_PROGRAM
    
            glDisable(GL_FRAGMENT_PROGRAM_ARB);

            #endif
        }
        
       // glClear(GL_COLOR_BUFFER_BIT);

#ifdef TILE_RENDER
    } while (trEndTile(tr));
#endif

#ifdef TIMING
    struct timeval newtv;
    gettimeofday(&newtv, NULL);

    printf("Time elapsed for Draw (ms): %d\n", newtv.tv_sec*1000 + newtv.tv_usec/1000 - tv.tv_sec*1000 - tv.tv_usec/1000);
#endif

    glClear(GL_COLOR_BUFFER_BIT);

    glutSwapBuffers();

    #ifdef SAGE_APP
        #ifndef TILE_RENDER
            glReadBuffer(GL_AUX0);
            //if(BYTES_PER_PIXEL == 3)
            //    glReadPixels(0, 0, total_width, total_height, GL_RGB, GL_UNSIGNED_BYTE, total_buffer);
            //else
                glReadPixels(0, 0, total_width, total_height, GL_RGBA, GL_UNSIGNED_BYTE, total_buffer);
        #endif
#ifdef TIMING
    gettimeofday(&tv, NULL);
#endif
        sageInf.swapBuffer(total_buffer);
#ifdef TIMING
    gettimeofday(&newtv, NULL);

    printf("Time elapsed for SAGE swapBuffer (ms): %d\n", newtv.tv_sec*1000 + newtv.tv_usec/1000 - tv.tv_sec*1000 - tv.tv_usec/1000);
#endif
    #endif


    

    
    glutPostRedisplay();
}

void ReshapeGL (int width, int height)    								// Reshape The Window When It's Moved Or Resized
{
    glViewport (0, 0, (GLsizei)(width), (GLsizei)(height));				// Reset The Current Viewport
    glMatrixMode (GL_PROJECTION);										// Select The Projection Matrix
    glLoadIdentity ();													// Reset The Projection Matrix
    //gluPerspective (45.0f, (GLfloat)(width)/(GLfloat)(height),			// Calculate The Aspect Ratio Of The Window
    //				0.1f, 100.0f);		
    glMatrixMode (GL_MODELVIEW);										// Select The Modelview Matrix
    glLoadIdentity ();													// Reset The Modelview Matrix

    glutPostRedisplay();
}


void Key(unsigned char key,int x,int y)
{
    Update(key,x,y);
    return ;
}

void OnIdle()
{
    glutPostRedisplay();
}


int main( int argc, char *argv[] )
{

    fprintf(stderr, "Amalgamator v4.0, compiled using multithreading and network sockets.\n");
    
    //the rest of initialization is done in Initialize(),
    //but we need these two numbers before creating the GLUT
    //window
    total_width = atoi(argv[2]);
    total_height = atoi(argv[3]);

    glutInit( &argc, argv );
    glutInitWindowPosition( 0, 0 );
#ifdef TILE_RENDER
    glutInitWindowSize( TILE_WIDTH, TILE_HEIGHT );
#else    
    glutInitWindowSize( total_width, total_height );
#endif

    //if(BYTES_PER_PIXEL == 3)
    //    glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH );
    //else
        glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH );


    glutCreateWindow("Amalgamator");//argv[0]);
    glutReshapeFunc( ReshapeGL );
    glutKeyboardFunc( Key );
    glutSpecialFunc( Update );
    glutDisplayFunc( Draw );
    glutIdleFunc( OnIdle );
    
    Initialize(argc, argv);
    
    int num_buffers = 0;
    glGetIntegerv(GL_AUX_BUFFERS, &num_buffers);
    printf("Number of buffers: %d.\n", num_buffers);
    
    pthread_t thread1, thread2;


    //CHANGE THESE IP ADDRESSES TO MATCH THOSE OF THE SENDERS


    int arg_pass[6];
    arg_pass[0] = 0;
    arg_pass[1] = 1234;
    arg_pass[2] = 137;
    arg_pass[3] = 110;
    arg_pass[4] = 119;
    arg_pass[5] = 188;

    pthread_create(&thread1, NULL, stream_thread_main, arg_pass);
   
    int arg_pass2[6];
    arg_pass2[0] = 1;
    arg_pass2[1] = 1234;
    arg_pass2[2] = 137;
    arg_pass2[3] = 110;
    arg_pass2[4] = 119;
    arg_pass2[5] = 204;

    pthread_create(&thread2, NULL, stream_thread_main, arg_pass2);
    //stream_thread_main(arg_pass);

       
    glutMainLoop();
    return 0;
}

void drawPlane(const Matrix4x4 &mTransform) 
{
    Vector3 p1 = Vector3( mTransform * Vector3(0, 0, 0) );
    Vector3 p2 = Vector3( mTransform * Vector3(individual_width, 0, 0) );
    Vector3 p3 = Vector3( mTransform * Vector3(individual_width, individual_height, 0) );
    Vector3 p4 = Vector3( mTransform * Vector3(0, individual_height, 0) );

    glColor4f(0,0,0,0);
    
    glBegin(GL_QUADS);
        
        glTexCoord2f(0.0f, 1.0f); glVertex3fv(p1);
        glTexCoord2f(1.0f, 1.0f); glVertex3fv(p2);
        glTexCoord2f(1.0f, 0.0f); glVertex3fv(p3);
        glTexCoord2f(0.0f, 0.0f); glVertex3fv(p4);
        
    glEnd();
}

// main function for recv threads
void *stream_thread_main(void *args)
{
    int thread_number = (int)((int*)args)[0];
    int port_number = (int)((int*)args)[1];

    int ip1 = (int)((int*)args)[2], ip2 = (int)((int*)args)[3], ip3 = (int)((int*)args)[4], ip4 = (int)((int*)args)[5];
    
    fprintf(stderr, "intialized thread number %d for port %d.\n", thread_number, port_number);


    int brate = 10000;
    
    // init new xvid interface object
    vvXviD * testx = new vvXviD(24.0f, 1, 31, brate, 250);

    // xvid encode buffers
    unsigned char * input;
    unsigned char * output;
    int esize, osize, fastmode = 0;

    // init encode buffers
    input = new unsigned char[individual_width*individual_height*3];
    output = new unsigned char[individual_width*individual_height*3];

    // Used for socket init
    int socketDescriptor;
    unsigned short int serverPort;
    struct sockaddr_in serverAddress;
    struct hostent * hostInfo;
    char buf[20];

    // format server address passed at start of thread
    snprintf(buf, 20, "%d.%d.%d.%d", ip1, ip2, ip3, ip4);
    printf("Thread %d is looking for input from ip %s.\n", thread_number, buf);

    // convert host name
    hostInfo = gethostbyname(buf);
    if (hostInfo == NULL) {
       fprintf(stderr, "problem converting host name");
       exit(1);
    }

    // port where server is listening
    serverPort = 1234;

    // establish socket
    socketDescriptor = socket(AF_INET, SOCK_STREAM, 0);
    if (socketDescriptor < 0 ) {
       fprintf(stderr, "cannot create socket");
       exit(1);
    } else 
       fprintf(stderr, "Thread %d socket created\n", thread_number);

//    int bdp;
//    bdp = 20000000;
//    setsockopt( socketDescriptor, SOL_SOCKET, SO_RCVBUF, (char *)&bdp, sizeof(bdp) );

    serverAddress.sin_family = hostInfo->h_addrtype;
    memcpy((char *) &serverAddress.sin_addr.s_addr,
            hostInfo->h_addr_list[0], hostInfo->h_length);
    serverAddress.sin_port = htons(serverPort);

    // connect to server
    if (connect(socketDescriptor,
        (struct sockaddr *) &serverAddress, sizeof(serverAddress)) < 0) {
       fprintf(stderr, "cannot connect");
       exit(1);
    } 

    PRINTF(printf("thread %d here: 1.\n", thread_number));

    // init xvid encoder and decoder to proper image size
    testx->create_enc(individual_width, individual_height);
    PRINTF(printf("thread %d here: 2.\n", thread_number));
    testx->create_dec(individual_width, individual_height);
  
    PRINTF(printf("Thread %d before socket set.\n", thread_number));

    // store socket id so other threads can find it
    socketfd[thread_number] = socketDescriptor;
    PRINTF(printf("Thread %d after the loop.\n", thread_number));
  
    // main recv loop
    while(1)
    {     
       PRINTF(printf("Thread %d entered main loop.\n", thread_number));

       // get word with size of next frame
       recv(socketDescriptor, &esize, 4, MSG_WAITALL);
       // if the size is zero, the sender sent a bitrate before the frame
       if(esize == 0)
       {
          PRINTF(printf("Thread %d. esize is zero! Attempting to receieve esize...\n", thread_number));
          int tempflag = 1;
          // get bitrate
          recv(socketDescriptor, &esize, 4, MSG_WAITALL);
          // store number in array
          streambr[thread_number] = esize;
          // see if all streams have a number stored
          for(int i = 0; i < number_of_streams; i++)
          {
             if( streambr[i] == 0 )
             {
                tempflag = 0;
                break;
             }
          }
          // if so, flag and call to have actual bitrate printed from the array
          if(brflag[0] == 0 && tempflag == 1)
          {
             brflag[0] = tempflag;
             Key('p', 0, 0);
          }
          // get size of next frame
          recv(socketDescriptor, &esize, 4, MSG_WAITALL);
          PRINTF(printf("Thread %d. Got esize: %d.\n", thread_number, esize));
       }

       // fastmode refers to sending the frame with no xvid compression
       // currently gets a few more fps, but takes much more bandwidth
       if (esize == individual_width*individual_height*3)
	  fastmode = 1;
       else
	  fastmode = 0;	

       PRINTF(printf("Thread %d. Attempting to retrieve one frame...\n", thread_number));

       // get frame from socket and read into correct buffer
       if(fastmode && BYTES_PER_PIXEL == 4)
       {
          recv(socketDescriptor, output, esize, MSG_WAITALL);
       } else if(fastmode) {
          recv(socketDescriptor, individual_buffers[thread_number], esize, MSG_WAITALL);
       } else {
	  recv(socketDescriptor, input, esize, MSG_WAITALL);	
       }

       // if RGB24
       if(BYTES_PER_PIXEL == 3)
       {
          while(pthread_mutex_lock(buffer_mutexes[thread_number]))
          {
             usleep(100);
             //fprintf(stderr, "Error! Unable to lock buffer mutex %d.\n", thread_number);
             //exit(1);
          }
          // decode frame into global buffer array
      	  if (!fastmode)
             testx->dec_frame(input, individual_buffers[thread_number], esize, &osize);
          while(pthread_mutex_unlock(buffer_mutexes[thread_number]))
          {
             usleep(100);
             //fprintf(stderr, "Error! Unable to unlock buffer mutex %d.\n", thread_number);
             //exit(1);
          }
       } else {
          while(pthread_mutex_lock(buffer_mutexes[thread_number]))
          {
             usleep(100);
             //fprintf(stderr, "Error! Unable to lock buffer mutex %d.\n", thread_number);
             //exit(1);
          }
          // decode frame into local buffer
	  if(!fastmode)
             testx->dec_frame(input, output, esize, &osize);
 
          //convert the RGB buffer in output to RGBA.
          int index, indexBPPd;
          //set the whole thing to 255. Really this is only so the alpha is 255, but I thought
          //it might be faster to do the whole thing at once as opposed to pixel by pixel when 
          //memcpy'ing the bytes from output over.
          PRINTF(printf("Thread %d. Memsetting the individual buffer. Buffer index: %d.\n", thread_number, thread_number));
          memset(individual_buffers[thread_number], 255, individual_width*individual_height*4);
          PRINTF(printf("Done memsetting. Now going to copy the buffer from the camera into the individual buffer....\n", thread_number));
          for(int i = 0; i < individual_height; i++)
          {
             for(int j = 0; j < individual_width; j++)
             {
                index = i*individual_width+j;
                indexBPPd = index*3;
                //printf("i, j: (%d, %d)\n", i, j);
                individual_buffers[thread_number][index*4] = output[indexBPPd];
                individual_buffers[thread_number][index*4+1] = output[indexBPPd+1];
                individual_buffers[thread_number][index*4+2] = output[indexBPPd+2];
                //memcpy((void*)individual_buffers[thread_number][index*4], (void*)output[indexBPPd], 3);
                //if(output[indexBPPd] == 0 && output[indexBPPd+1] == 0 && output[indexBPPd+2] == 0)
                //    
             }
          }

          PRINTF(printf("Done.\n"));



          //printf("osize: %d. esize: %d.\n", osize, esize);
          //memcpy(individual_buffers[thread_number], output, individual_width*individual_height*3);
          while(pthread_mutex_unlock(buffer_mutexes[thread_number]))
          {
             usleep(100);
             //fprintf(stderr, "Error! Unable to unlock buffer mutex %d.\n", thread_number);
             //exit(1);
          }
       }
    
       PRINTF(printf("Thread %d. Success!\n", thread_number));
    }
}
